/*
PLProtocol.cpp
Written by Walter O. Krawec

Copyright (c) 2016 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef PL_PROTOCOL_H_
#define PL_PROTOCOL_H_

#include "../Quantum/Quantum.h"
#include "../QKD.h"

class PLProtocol : public QKDEvolve::QKD
{
public:
	virtual ~PLProtocol()
	{
	}

	virtual bool setup()
	{
		return true;
	}

	virtual void printDebugInfo(std::ostream& f, bool printRate)
	{
	};

	virtual int getESystemDimension(std::vector <int>& dim)
	{
		dim = dimE;
		return 1;

		dim.resize(3);
		dim[0] = 2;
		dim[1] = 1;
		dim[2] = 4;


		dim[1] = 4;
		dim[2] = 64;
		return 1;
	}

	virtual bool setup(quantum::DensityList& _primary, std::list <quantum::DensityList>& _qberList, std::list <quantum::DensityList>& _statList, std::list <quantum::DensityList>& _printList, std::vector <int> &_dimE)
	{
		dimE = _dimE;
		primary = _primary;
		primary.trace(1, primaryAE);
		primaryAE.trace(0, primaryE);

		primaryAENOT = primaryAE;
		primaryAENOT.NOT(0);
		qberList = _qberList;

		statList = _statList;

		printList = _printList;

		return true;
	}

	virtual double computeQBER(std::vector <quantum::Basis>& B, std::vector <double>& qberDebugList)
	{
		bool balancedMode = false;
		if(qberList.empty())
			return 0;

		if(qberDebugList.empty())
			qberDebugList.resize(qberList.size()+statList.size()+printList.size());

		std::list <quantum::DensityList>::iterator Iter;
		double Q = 0;

		int index=0;
		for(Iter = qberList.begin(); Iter != qberList.end(); ++Iter)
		{
			double p01 = Iter->calculatePr(0, 0, 1, 1, B);
			double p10 = Iter->calculatePr(0, 1, 1, 0, B);

			qberDebugList[index] = (p01+p10);
			++index;

			if(balancedMode)
			{
				// balanced mode:
				double p = 0;
				if(p10 > p01)
					p = p10;
				else
					p = p01;

				if(2*p > Q)
					Q = 2*p;
			}
			else
			{
				//regular mode:
				if(p01+p10 > Q)
					Q = p01+p10;
			}
		}

		for(Iter = statList.begin(); Iter != statList.end(); ++Iter)
		{
			double p01 = Iter->calculatePr(0, 0, 1, 1, B);
			double p10 = Iter->calculatePr(0, 1, 1, 0, B);

			qberDebugList[index] = (p01+p10);
			++index;

			double dist = p01+p10;
			dist = fabs(dist-.5);
			if(dist > Q)
				Q = dist;
		}

		for(Iter = printList.begin(); Iter != printList.end(); ++Iter)
		{
			double p01 = Iter->calculatePr(0, 0, 1, 1, B);
			double p10 = Iter->calculatePr(0, 1, 1, 0, B);

			qberDebugList[index] = (p01+p10);
			++index;
		}

		return Q;
	}

	double computeKeyRate(std::vector <quantum::Basis>& B, double q)
	{
		//double qSave = q;
		//q = .5;
		algebra::mat AE0, AE1, E;

		primaryAE.computeDensityMatFast(B, AE0);
		primaryAENOT.computeDensityMatFast(B, AE1);
		primaryE.computeDensityMatFast(B, E);

		Complex c = 1-q;

		AE0.multiplyScalar(c);
		c = q;
		AE1.multiplyScalar(c);

		AE0.add(&AE1);
		double s1 = primaryAE.entropy(AE0);
		double s2 = primaryE.entropy(E);

		//AE0.print(std::cout, true);

		//q = qSave;

		double p00 = primary.calculatePr(0, 0, 1, 0, B);
		double p11 = primary.calculatePr(0, 1, 1, 1, B);
		double p01 = primary.calculatePr(0, 0, 1, 1, B);
		double p10 = primary.calculatePr(0, 1, 1, 0, B);

		double newp00 = (1-q)*p00 + q*p10;
		double newp11 = (1-q)*p11 + q*p01;
		double newp01 = (1-q)*p01 + q*p11;
		double newp10 = (1-q)*p10 + q*p00;

		p00 = newp00;
		p11 = newp11;
		p01 = newp01;
		p10 = newp10;

		double p0 = p00+p10;
		double p1 = p01+p11;

		double test = p00+p11+p01+p10;

		test = p0+p1;

		double H = 0;
		if(p00 > 0.00000001)
			H += (p00*log(p0/p00));
		if(p11 > 0.00000001)
			H += (p11*log(p1/p11));
		if(p01 > 0.00000001)
			H += (p01*log(p1/p01));
		if(p10 > 0.00000001)
			H += (p10*log(p0/p10));

		H = H / log(2.0);

		double test2 = s2-s1;
		double Q2 = p01+p10;
		return s1 - s2 - H;
	}
	virtual double computeKeyRate(std::vector <quantum::Basis>& B, bool printDebug)
	{
		double q = 0;
		double deltaQ = .01;
		double bestR = -100;
		while(q <= .4+deltaQ/2.0) // .1, .25, and .4
		{
			if(q > .5)
				break;
			double r = computeKeyRate(B, q);
			if(r > bestR)
				bestR = r;
			q += deltaQ;
			break;				// remove to enable preprocessing
		}
		return bestR;

		algebra::mat AE, E;

		primaryAE.computeDensityMatFast(B, AE);
		primaryE.computeDensityMatFast(B, E);

		double s1 = primaryAE.entropy(AE);
		double s2 = primaryE.entropy(E);

		double p00 = primary.calculatePr(0, 0, 1, 0, B);
		double p11 = primary.calculatePr(0, 1, 1, 1, B);
		double p01 = primary.calculatePr(0, 0, 1, 1, B);
		double p10 = primary.calculatePr(0, 1, 1, 0, B);

		double p0 = p00+p10;
		double p1 = p01+p11;

		double H = 0;
		if(p00 > 0.00000001)
			H += (p00*log(p0/p00));
		if(p11 > 0.00000001)
			H += (p11*log(p1/p11));
		if(p01 > 0.00000001)
			H += (p01*log(p1/p01));
		if(p10 > 0.00000001)
			H += (p10*log(p0/p10));

		H = H / log(2.0);

		return s1 - s2 - H;
	}

	virtual void testSolution(std::vector <quantum::Basis>& B)
	{
		algebra::mat U1, U2, temp;

		U1.create(8);

		std::vector <double> Pr;

		int index = 0;
		for(int c=0; c<8; ++c)
		{
			for(int r=0; r<8; ++r)
			{
				U1.setDirect(r, c, B[0].B[index][0]);

				Complex c2 = B[0].B[index][0];
				algebra::ComplexHelper::sq(c2);
				Pr.push_back(creal(c2));
				++index;
			}
		}

		U2.create(512, 8);
	//	index = 64;
		index = 0;
		for(int c=0; c<8; ++ c)
		{
			for(int r=0; r<8; ++r)
			{
				for(int j=0; j<(int)B[1].B[index].size(); ++j)
				{
					U2.setDirect(64*r+j, c, B[1].B[index+r][j]);
				}
			}
			index += 64;
		}

		U2.transpose(U1);
		U1.multiply(U2, temp);

		std::ofstream f;
		f.open("mat1.txt");

		for(int i=0; i<Pr.size(); ++i)
			f << Pr[i] << "\n";

		temp.print(f, true);

		f.close();

		double sum = 0;
		for(int i=0; i<temp.getRows(); ++i)
		  sum += creal(temp.getDirect(i,i));

		std::cout << "SUM = " << sum << "\n";
	}

private:
	quantum::DensityList primary, primaryAE, primaryAENOT, primaryE;
	std::list <quantum::DensityList> qberList, statList, printList;
	std::vector <int> dimE;
};

#endif
