/*
PL.cpp
Written by Walter O. Krawec

Copyright (c) 2016 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "PL.h"
#include "Text.h"

namespace pl
{
	bool PL::constructDL(const std::string& filename, quantum::DensityList& primary, std::list <quantum::DensityList>& qberDL, std::list <quantum::DensityList>& statList, std::list <quantum::DensityList>& printList, std::vector <int>& dimE)
	{
		dimE.clear();
		bool finishedDimE = false;
		bool firstDimE = true;
		quantum::DensityList output;

		wok::Text text, text2, textInner, textCommand;
		text.addCollection(wok::Text::Pair('(', ')'));
		text.addCollection(wok::Text::Pair('{', '}'));
		text.addDelimiter(',');
		text.addDelimiter(' ');

		text.addIgnore('\r');
		text.addIgnore('\n');

		text2.addCollection(wok::Text::Pair('(', ')'));
		text2.addDelimiter(',');
		text2.addDelimiter(':');
		text2.addDelimiter('=');
		text2.addIgnore(' ');

		text2.addIgnore('\r');
		text2.addIgnore('\n');


		textCommand.addCollection(wok::Text::Pair('(', ')'));
		textCommand.addDelimiter(',');
		textCommand.addDelimiter(' ');

		textCommand.addIgnore('\r');
		textCommand.addIgnore('\n');


		textInner.addDelimiter(';');

		textInner.addIgnore('\r');
		textInner.addIgnore('\n');


		std::ifstream f;
		f.open(filename.c_str());
		if(!f.is_open())
			return false;

		int bufferSize = 1024;
		char* buffer;
		buffer = new char[bufferSize+10];

		std::vector <Pair <int> > subspaceNames;
		std::vector <Pair <algebra::mat> > definedOps;
		std::list <quantum::KetBra> baseDL, tempDL, primaryWithBaseDL;

		while(!f.eof())
		{
			f.getline(buffer, bufferSize);
			buffer[bufferSize-1] = 0;

			text.parse(buffer);

			if(text[0] == "create" && text[1] == "space")
			{
				output.density.clear();
				output.dimension.clear();
				text2.parse(text[2]);
				subspaceNames.clear();
				for(int i=0; i<text2.size(); i+=2)
				{
					Pair <int> p;
					p.name = text2[i];
					p.value = atoi(text2[i+1].c_str());

					//std::cout << "NAME = '" << p.name << "'\n\tV = " << p.value << "\n\n";

					subspaceNames.push_back(p);
					output.dimension.push_back(p.value);
				}

				quantum::KetBra kb(output.dimension.size());
				kb.p = 1;

				output.density.push_back(kb);
			}
			else if(text[0] == "save" && text[1] == "as")
			{
				finishedDimE = true;
				if(text[2] == "primary")
					primary = output;
				else if(text[2] == "QBER")
					qberDL.push_back(output);
				else if(text[2] == "stat")
					statList.push_back(output);
				else if(text[2] == "print")
					printList.push_back(output);
			}
			else if(text[0] == "define")
			{
				Pair <algebra::mat> p;
				p.name = text[1];
				if(text[3] == "QFT")
				{
					int dim = atoi(text[4].c_str());
					p.value.create(dim);
					for(int r=0; r<dim; ++r)
					{
						for(int c=0; c<dim; ++c)
						{
							Complex rt;
							double pi = 3.1415926535897932384626433832795;
							rt = algebra::ComplexHelper::construct(2.0*pi/(double)(dim));
							p.value(r,c) = algebra::ComplexHelper::power(rt, r*c);

							//std::cout << "\t(" << r << ", "  << c << "): rt = " << creal(rt) << "+" << cimag(rt) << "i\n";
						}
					}

					Complex scale;
					scale = 1.0 / sqrt((double)dim);

					p.value.multiplyScalar(scale);
				}
				else if(text[3] == "ChangeBasis")
				{
					p.value.create(2);
					p.value.zero();
					if(text[4] == "Y")
					{
						p.value(0,0) = 1.0/sqrt(2.0);
						p.value(1,0) = 1.0/sqrt(2.0)*I;
						p.value(0,1) = 1.0/sqrt(2.0);
						p.value(1,1) = -1.0/sqrt(2.0);
					}
				}
				else if(text[3] == "Pauli")
				{
					p.value.create(2);
					p.value.zero();
					if(text[4] == "X")
					{
						p.value(0,1) = 1;
						p.value(1,0) = 1;
					}
					else if(text[4] == "Y")
					{
						p.value(0,1) = -1*I;
						p.value(1,0) = 1*I;
					}
					else if(text[4] == "Z")
					{
						p.value(0,0) = 1;
						p.value(1,1) = -1;
					}
				}

				definedOps.push_back(p);
			}
			else if(text[0] == "conditional" && text[1] == "set")
			{
				std::vector <int> targetSpace, condSpace;
				std::vector <int> targetVal, condVal;
				text2.parse(text[2]);

				for(int i=0; i<(int)text2.size(); i+=2)
				{
					std::string name = text2[i];
					bool found = false;
					int space=0;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";
					targetSpace.push_back(space);
					targetVal.push_back(atoi(text2[i+1].c_str()));
				}

				text2.parse(text[4]);
				for(int i=0; i<(int)text2.size(); i+=2)
				{
					std::string name = text2[i];
					bool found = false;
					int space=0;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";
					condSpace.push_back(space);
					condVal.push_back(atoi(text2[i+1].c_str()));
				}

				output.conditionalSet(targetSpace, targetVal, condSpace, condVal);
			}
			else if(text[0] == "project")
			{
				std::string name = text[1];
				bool found = false;
				int space=0;
				for(int i=0; i<subspaceNames.size(); ++i)
				{
					if(subspaceNames[i].name == name)
					{
						found = true;
						space = i;
						break;
					}
				}
				if(!found)
					std::cout << "ERROR: subspace '" << name << "' not defined.\n";

				int state = atoi(text[3].c_str());
				output.project(space, state);
			}
			else if(text[0] == "swap")
			{
				std::string name = text[1];
				bool found = false;
				int space1=0, space2=0;
				for(int i=0; i<subspaceNames.size(); ++i)
				{
					if(subspaceNames[i].name == name)
					{
						found = true;
						space1 = i;
						break;
					}
				}
				if(!found)
					std::cout << "ERROR: subspace '" << name << "' not defined.\n";

				name = text[3];
				found = false;
				for(int i=0; i<subspaceNames.size(); ++i)
				{
					if(subspaceNames[i].name == name)
					{
						found = true;
						space2 = i;
						break;
					}
				}
				if(!found)
					std::cout << "ERROR: subspace '" << name << "' not defined.\n";
				output.swap(space1, space2);
			}
			else if(text[0] == "measure")
			{
				std::string name = text[1];

				int space = 0, saveSpace = 1;
				bool found = false;
				for(int i=0; i<subspaceNames.size(); ++i)
				{
					if(subspaceNames[i].name == name)
					{
						found = true;
						space = i;
						break;
					}
				}
				if(!found){
				  std::cout << "ERROR2\n";
					std::cout << "ERROR: subspace '" << name << "' not defined.\n";
				}

				name = text[4];

				found = false;
				for(int i=0; i<subspaceNames.size(); ++i)
				{
					if(subspaceNames[i].name == name)
					{
						found = true;
						saveSpace = i;
						break;
					}
				}
				if(!found)
				  {
				    std::cout << "ERROR3\n";
					std::cout << "ERROR: subspace '" << name << "' not defined.\n";
				  }

				output.measureAndSave(space, saveSpace);
			}
			else if(text[0] == "trace" && text[1] == "out")
			{
				std::string spaceName = text[2];
				int space = 0;
				algebra::mat op;
				bool found = false;
				for(int i=0; i<subspaceNames.size(); ++i)
				{
					if(subspaceNames[i].name == spaceName)
					{
						found = true;
						space = i;
						break;
					}
				}
				if(!found)
					std::cout << "ERROR: subspace '" << spaceName << "' not defined.\n";

				output.trace(space);
			}
			else if(text[0] == "apply" && text[1] == "op")
			{
				std::string opName = text[2];
				int space = 0;
				std::string name = text[4];
				algebra::mat op;
				bool found = false;
				for(int i=0; i<subspaceNames.size(); ++i)
				{
					if(subspaceNames[i].name == name)
					{
						found = true;
						space = i;
						break;
					}
				}
				if(!found)
					std::cout << "ERROR: subspace '" << name << "' not defined.\n";

				if(opName == "NOT")
					output.NOT(space);
				else
				{
					found = false;
					for(int i=0; i<(int)definedOps.size(); ++i)
					{
						if(definedOps[i].name == opName)
						{
							found = true;
							op = definedOps[i].value;
							break;
						}
					}
					if(!found)
					{
						std::cout << "ERROR: op '" << name << "' not defined.\n";
					}
					else
					{
						output.applyOp(space, op);
					}
				}
			}
			else if(text[0] == "apply" && text[1] == "conditional" && text[2] == "op")
			{
				std::string spaceName = text[5];
				int space = 0;
				std::string name = text[3];
				algebra::mat op;
				bool found = false;
				for(int i=0; i<subspaceNames.size(); ++i)
				{
					if(subspaceNames[i].name == spaceName)
					{
						found = true;
						space = i;
						break;
					}
				}
				if(!found)
					std::cout << "ERROR: subspace '" << spaceName << "' not defined.\n";
				found = false;
				for(int i=0; i<(int)definedOps.size(); ++i)
				{
					if(definedOps[i].name == name)
					{
						found = true;
						op = definedOps[i].value;
						break;
					}
				}
				if(!found)
				{
					std::cout << "ERROR: op '" << name << "' not defined.\n";
				}
				else
				{
					text2.parse(text[7]);
					name = text2[0];
					int condSpace = 0, cond = atoi(text2[1].c_str());
					found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							condSpace = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";
					output.applyConditionalOp(condSpace, cond, space, op);
				}
			}
			else if(text[0] == "attack")
			{
				text2.parse(text[1]);
				int smallSpace, largeSpace;
				int space = 0;
				for(int i=0; i<(int)text2.size(); ++i)
				{
					std::string name = text2[i];
					bool found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";
					if(i == 0 || space > largeSpace)
						largeSpace = space;
					if(i == 0 || space < smallSpace)
						smallSpace = space;
				}

				if(!finishedDimE)
				{
					int dim = 1;
					if(firstDimE)
					{
						// first means this is the transit dim
						for(int i=smallSpace; i<=largeSpace; ++i)
							dim *= output.dimension[i];

						dimE.push_back(dim);
						firstDimE = false;
					}
					dim = output.dimension[largeSpace+1];
					dimE.push_back(dim);
				}

				output.applyOp(smallSpace, largeSpace+1);
			}
			else if(text[0] == "endwith")
			{
				output.density = tempDL;
			}
			else if(text[0] == "debug" && text[1] == "print")
			{
				output.print(std::cout);
			}
			else if((text[0] == "with" || text[0] == "elsewith") && text[1] == "prob")
			{
				double pr = atof(text[2].c_str());
				if(text[0] == "with")
				{
					baseDL = output.density;
					tempDL.clear();
				}
				if(text[3] == "prepare")
				{
					text2.parse(text[4]);
					std::list <quantum::KetBra>::iterator Iter;
					for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						kb.p *= pr;

						for(int i=0; i<text2.size(); i+=2)
						{
							std::string name = text2[i];
							int value = atoi(text2[i+1].c_str());
							
							for(int j=0; j<(int)subspaceNames.size(); ++j)
							{
								if(subspaceNames[j].name == name)
								{
									kb.ket[j] = value;
									kb.bra[j] = value;
								}
							}
						}

						tempDL.push_back(kb);
					}
				}
				else if(text[3] == "swap")
				{
					std::string name = text[4];
					
					int space1 = 0, space2 = 1;
					bool found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space1 = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";

					name = text[6];
					found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space2 = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";

					std::list <quantum::KetBra>::iterator Iter;
					quantum::DensityList tempDL2;
					tempDL2.dimension = output.dimension;
					for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						kb.p *= pr;

						tempDL2.density.push_back(kb);
					}

					tempDL2.swap(space1, space2);
					for(Iter = tempDL2.density.begin(); Iter != tempDL2.density.end(); ++Iter)
						tempDL.push_back(*Iter);
				}
				else if(text[3] == "nothing")
				{
					std::list <quantum::KetBra>::iterator Iter;
					for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						kb.p *= pr;

						tempDL.push_back(kb);
					}
				}
				else if(text[3] == "measure")
				{
					std::list <quantum::KetBra>::iterator Iter;
					quantum::DensityList tempDL2;
					tempDL2.dimension = output.dimension;
					for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						kb.p *= pr;

						tempDL2.density.push_back(kb);
					}

					std::string name = text[4];
					
					int space = 0, saveSpace = 1;
					bool found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";

					name = text[7];
					found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							saveSpace = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";

					tempDL2.measureAndSave(space, saveSpace);
					for(Iter = tempDL2.density.begin(); Iter != tempDL2.density.end(); ++Iter)
						tempDL.push_back(*Iter);
				}
			}
			else if((text[0] == "with-set" || text[0] == "elsewith-set") && text[1] == "prob")
			{
				double pr = atof(text[2].c_str());
				if(text[0] == "with-set")
				{
					baseDL = output.density;
					tempDL.clear();
				}
				if(text[3] == "prepare")
				{
					text2.parse(text[4]);
					std::list <quantum::KetBra>::iterator Iter;
					for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						kb.p *= pr;

						for(int i=0; i<text2.size(); i+=2)
						{
							std::string name = text2[i];
							int value = atoi(text2[i+1].c_str());

							for(int j=0; j<(int)subspaceNames.size(); ++j)
							{
								if(subspaceNames[j].name == name)
								{
									kb.ket[j] = value;
									kb.bra[j] = value;
								}
							}
						}

						tempDL.push_back(kb);
					}
				}
				else if(text[4] == "swap")
				{
					std::string name = text[5];
					int space1 = 0, space2 = 1;
					bool found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space1 = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";

					name = text[7];
					found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space2 = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";

					std::list <quantum::KetBra>::iterator Iter;
					quantum::DensityList tempDL2;
					tempDL2.dimension = output.dimension;
					for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						kb.p *= pr;

						tempDL2.density.push_back(kb);
					}

					// NOW PERFORM SET
					text2.parse(text[3]);
					//std::list <quantum::KetBra>::iterator Iter;
					for(Iter = tempDL2.density.begin(); Iter != tempDL2.density.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						for(int i=0; i<text2.size(); i+=2)
						{
							std::string name = text2[i];
							int value = atoi(text2[i+1].c_str());

							for(int j=0; j<(int)subspaceNames.size(); ++j)
							{
								if(subspaceNames[j].name == name)
								{
									kb.ket[j] = value;
									kb.bra[j] = value;
								}
							}
						}

						*Iter = kb;
					}

					tempDL2.swap(space1, space2);
					for(Iter = tempDL2.density.begin(); Iter != tempDL2.density.end(); ++Iter)
						tempDL.push_back(*Iter);

					
				}
				else if(text[4] == "nothing")
				{
					std::list <quantum::KetBra>::iterator Iter;
					std::list <quantum::KetBra> tempDL2;
					for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						kb.p *= pr;

						tempDL2.push_back(kb);
					}

					// NOW PERFORM SET
					text2.parse(text[3]);
					//std::list <quantum::KetBra>::iterator Iter;
					for(Iter = tempDL2.begin(); Iter != tempDL2.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						for(int i=0; i<text2.size(); i+=2)
						{
							std::string name = text2[i];
							int value = atoi(text2[i+1].c_str());

							for(int j=0; j<(int)subspaceNames.size(); ++j)
							{
								if(subspaceNames[j].name == name)
								{
									kb.ket[j] = value;
									kb.bra[j] = value;
								}
							}
						}

						*Iter = kb;
					}

					for(Iter = tempDL2.begin(); Iter != tempDL2.end(); ++Iter)
						tempDL.push_back(*Iter);
				}
				else if(text[4] == "measure")
				{
					std::list <quantum::KetBra>::iterator Iter;
					quantum::DensityList tempDL2;
					tempDL2.dimension = output.dimension;
					for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						kb.p *= pr;

						tempDL2.density.push_back(kb);
					}

					std::string name = text[5];

					int space = 0, saveSpace = 1;
					bool found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							space = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";

					name = text[8];
					found = false;
					for(int i=0; i<subspaceNames.size(); ++i)
					{
						if(subspaceNames[i].name == name)
						{
							found = true;
							saveSpace = i;
							break;
						}
					}
					if(!found)
						std::cout << "ERROR: subspace '" << name << "' not defined.\n";

					// NOW PERFORM SET
					text2.parse(text[3]);
					//std::list <quantum::KetBra>::iterator Iter;
					for(Iter = tempDL.begin(); Iter != tempDL.end(); ++Iter)
					{
						quantum::KetBra kb = *Iter;
						for(int i=0; i<text2.size(); i+=2)
						{
							std::string name = text2[i];
							int value = atoi(text2[i+1].c_str());

							for(int j=0; j<(int)subspaceNames.size(); ++j)
							{
								if(subspaceNames[j].name == name)
								{
									kb.ket[j] = value;
									kb.bra[j] = value;
								}
							}
						}

						*Iter = kb;
					}

					tempDL2.measureAndSave(space, saveSpace);
					for(Iter = tempDL2.density.begin(); Iter != tempDL2.density.end(); ++Iter)
						tempDL.push_back(*Iter);

					
				}
			}
			else if((text[0] == "with-prob" || text[0] == "elsewith-prob"))
			{
				double pr = atof(text[1].c_str());
				if(text[0] == "with-prob")
				{
					baseDL = output.density; // this is what we're building each step of the "with" block
					primaryWithBaseDL = output.density;  // this is the outer most original DL
					tempDL.clear();  // this is the final result
				}
				else
				{
					baseDL = primaryWithBaseDL;
				}
				// text[2] should be the collection {cmd1;cmd2;...}
				textInner.parse(text[2]);
				for(int i=0; i<textInner.size(); ++i)
				{
					textCommand.parse(textInner[i]);
					if(textCommand[0] == "prepare")
					{
						text2.parse(textCommand[1]);
						std::list <quantum::KetBra>::iterator Iter;
						for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
						{
							//quantum::KetBra kb = *Iter;
							//kb.p.real *= pr;
							//kb.p.imag *= pr;
							for(int i=0; i<text2.size(); i+=2)
							{
								std::string name = text2[i];
								int value = atoi(text2[i+1].c_str());

								for(int j=0; j<(int)subspaceNames.size(); ++j)
								{
									if(subspaceNames[j].name == name)
									{
										if(Iter->ket[j] != Iter->bra[j])
											std::cout << "Warning: possible error with prepare statement: system not |0><0|\n";
										Iter->ket[j] = value;
										Iter->bra[j] = value;
									}
								}
							}
						}
					}
					else if(textCommand[0] == "do-nothing")
					{
					}
					else if(textCommand[0] == "measure")
					{
						std::list <quantum::KetBra>::iterator Iter;
						quantum::DensityList tempDL2;
						tempDL2.dimension = output.dimension;
						for(Iter = baseDL.begin(); Iter != baseDL.end(); ++Iter)
						{
							quantum::KetBra kb = *Iter;
							//kb.p.real *= pr;
							//kb.p.imag *= pr;
							tempDL2.density.push_back(kb);
						}

						std::string name = textCommand[1];

						int space = 0, saveSpace = 1;
						bool found = false;
						for(int i=0; i<subspaceNames.size(); ++i)
						{
							if(subspaceNames[i].name == name)
							{
								found = true;
								space = i;
								break;
							}
						}
						if(!found)
							std::cout << "ERROR: subspace '" << name << "' not defined.\n";

						name = textCommand[4];
						found = false;
						for(int i=0; i<subspaceNames.size(); ++i)
						{
							if(subspaceNames[i].name == name)
							{
								found = true;
								saveSpace = i;
								break;
							}
						}
						if(!found)
							std::cout << "ERROR: subspace '" << name << "' not defined.\n";

						tempDL2.measureAndSave(space, saveSpace);
						baseDL.clear();
						for(Iter = tempDL2.density.begin(); Iter != tempDL2.density.end(); ++Iter)
							baseDL.push_back(*Iter);
					}
				}

				// now that all commands have been run, add "baseDL" to "tempDL"
				std::list <quantum::KetBra>::iterator Iter;
				for(Iter=baseDL.begin(); Iter != baseDL.end(); ++Iter)
				{
					quantum::KetBra kb = *Iter;
					kb.p *= pr;

					tempDL.push_back(kb);
				}
			}
		}

		delete[] buffer;
		f.close();
	}
}
