/*
QKD.h
Written by Walter O. Krawec

Copyright (c) 2016 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef QKD_H_
#define QKD_H_

#include "Quantum/Quantum.h"

#ifdef WIN32
#include <conio.h>
#endif

#include <fstream>

namespace QKDEvolve
{
class QKD
{
public:
	virtual ~QKD()
	{
	}

	// should setup "system"
	virtual bool setup()=0;

	// should initialize "B" so that all var. bases are configured to the correct size
	//virtual void setupNewBasis(std::vector <quantum::Basis>& B, std::vector <quantum::Basis>& Bsmall)=0;

	// should setup dim so that dim[0] = dim H_T, dim[i] = dim H_{E,i}
	// also should return index of subspace for H_{E,1}
	virtual int getESystemDimension(std::vector <int>& dim)=0;

	// computes the QBER induced by "B"
	virtual double computeQBER(std::vector <quantum::Basis>& B, std::vector <double>& qberDebugList)=0;

	// computes the key rate induced by "B"
	virtual double computeKeyRate(std::vector <quantum::Basis>& B, bool printDebug)=0;

	// computes S(E) induced by "B"
	virtual double computeE(std::vector <quantum::Basis>& B){return 0;};

	virtual void computeDifferentH(double& H1, double& H2, std::vector <quantum::Basis>& B){};

	virtual void printDebugInfo(std::ostream& f, bool printRate){};

	virtual void testSolution(std::vector <quantum::Basis>& B) {};

	double entropy(double x)
	{
		if(x <= 0.00000001)
			return 0;
		if(x >= 1-.00000001)
			return 0;

		double h = -x*log(x) - (1-x)*log(1-x);
		h = h / log(2.0);

		return h;
	}

protected:
	quantum::DensityList system;
};

struct CS
{
	CS() : H1(0), H2(0) {};
	double rateFit, qberFit, EFit, Q, fitness, H1, H2;
	std::vector <int> dim, prodDim;
	std::vector <double> qberList;

	//std::vector <quantum::Basis> B, Bsmall, orig;
	// G contains non-normalized, non-orthogonal basis states (solution representation)
	// Next, these are orthgonalized and stored in "F"
	// Finally, "E" are "short" versions of "F": E[i].B[j] is the state |e_j> for H_{E,i}
	std::vector <quantum::Basis> G, F, E;

	void saveLocalDim1(const char* filename)
	{
		std::ofstream f;
		f.open(filename);

		int i = 0;

		for(unsigned int j=0; j<E[i].B.size(); ++j)
		{
			//f << "eu" << i << "u" << j << " = {";
			for(unsigned int k=0; k<E[i].B[j].size(); ++k)
			{
			  f << "e" << j << ".setDirect(" << k << ", 0, " << creal(E[i].B[j][k]) << " , " << cimag(E[i].B[j][k]) << ");\n";
			}
			f << "\n";
		}

		f.close();
	}
	void saveMathematica(const char* filename)
	{
		std::ofstream f;
		f.open(filename);

		for(unsigned int i=0; i<E.size(); ++i)
		{
			if(E[i].B.empty()) continue;

			for(unsigned int j=0; j<E[i].B.size(); ++j)
			{
				f << "eu" << i << "u" << j << " = {";
				for(unsigned int k=0; k<E[i].B[j].size(); ++k)
				{
					if(k != 0)
						f << ",";
					f << "{" << creal(E[i].B[j][k]) << " + " << cimag(E[i].B[j][k]) << "*I}";
				}
				f << "};\n";
			}
			f << "\n";
		}

		f.close();
	}

	bool operator< (const CS& other)
	{
		if(fitness < other.fitness)
			return true;
		return false;

		if(rateFit < 0 && other.rateFit < 0)
		{
			if(Q < other.Q)
				return true;
			return false;
		}
		if(rateFit < other.rateFit)
			return true;
		return false;

		if(fitness < other.fitness)
			return true;
		return false;

		if(qberFit < .0005 && other.qberFit < .0005)
		{
			if(rateFit < other.rateFit)
				return true;
			return false;
		}
		if(qberFit < other.qberFit)
			return true;
		return false;
	}

	double randFloat()
	{
		int r = rand()%RAND_MAX;
		double f = r;
		f = f / (double)RAND_MAX;
		return 4.0*f-2.0;
	}

	void calculateFitness(QKD* protocol, double targetQBER)
	{
		bool useRate = true;

		rateFit = 0;
		if(useRate)
			rateFit = protocol->computeKeyRate(E, false);

		EFit = protocol->computeE(E);
		qberFit = protocol->computeQBER(E, qberList);
		Q = qberFit;
		qberFit = (qberFit-targetQBER/2.0)*(qberFit-targetQBER/2.0);
		double saveRate = rateFit;
		double targetRate = -.01;
		//targetRate = -1;
		//targetRate = -.5;
		rateFit = (rateFit-targetRate)*(rateFit-targetRate);
		//rateFit = saveRate;
		
		double p = .5;//.99;//.98;
		//p = .98;
		//p = .9;
		//p = .02;  // .1 needed for postprocessing of .2; .02 for .4; .08 for .25
		if(useRate)
			fitness = p*qberFit + (1-p)*(rateFit);
		else
			fitness = qberFit;

		//fitness = saveRate*qberFit;

		rateFit = saveRate;
		//if(Q <= .11)
		//	fitness = rateFit;

		//fitness = qberFit;

		protocol->computeDifferentH(H1, H2, E);
	}

	void setup(QKD* protocol, double largeScale, double smallScale)
	{
		if(!protocol) return;

		// setupNewBasis should first initialize G so that G.size() = # of subspaces used (even those that are not H_E)
		// Next, for each H_{E,i}, which is the j(i)'th subspace of the total system, set G[j(i)].size() = d_i (the dim H_{E,i})

		// dim[i] = d_i where d_0 = dim H_T, d_i = dim H_{E,i}
		protocol->getESystemDimension(dim);

		if(dim.size() == 0)
			throw std::string("CS::setup error: dim.size() == 0");

		std::vector <int> saveDim = dim, prodDim2;
		dim[1] = dim[0]*dim[0];
		if(dim.size() > 2)
			dim[2] = dim[1]*dim[1]*dim[1];

		int prod = 1, prod2 = 1;
		for(int i=0; i<(int)dim.size(); ++i)
		{
			prodDim.push_back(prod);
			prod *= dim[i];

			prodDim2.push_back(prod2);
			prod2 *= saveDim[i];
		}

		G.resize(dim.size()-1);
		E.resize(dim.size()-1);
		for(unsigned int r=0; r<dim.size()-1; ++r)
		{
			// G[r].B[i] = (g^r_{i,1}, g^r_{i,2}, ..., g^r_{i,D_r})
			// each g^r_{i,j} \in C^dim[r]
			prod = prodDim2[r+1] * saveDim[r+1];
			G[r].B.resize(prodDim2[r+1]);
			for(int i=0; i<prodDim2[r+1]; ++i)
				G[r].B[i].resize(prod);

			if(r == 0)
				E[r].B.resize(prodDim2[r+1]*prodDim2[r+1]);
			else
				E[r].B.resize(prodDim2[r+1]*prodDim2[r]*E[r-1].B.size());
			for(int i=0; i<(int)E[r].B.size(); ++i)
				E[r].B[i].resize(saveDim[r+1]);
		}

		//float largeScale = .04;
		//float smallScale = .04;

		//largeScale = .1;
		//smallScale = .07;

		//double saveLarge = largeScale;
		//double saveSmall = smallScale;
		for(unsigned int i=0; i<G.size(); ++i)
		{
			// put this back to use identity attack - useful for checking correctness of protocol
			/**if(i == 0)
			{
				smallScale = 0;
				largeScale = 0;
			}
			else
			{
				smallScale = 0;
				largeScale = 0;
			}**/
			int index = 0;
			for(unsigned int j=0; j<G[i].B.size(); ++j)
			{
				for(unsigned int k=0; k<G[i].B[j].size(); ++k)
				{
					if(k >= E[i].B[j].size()*index && k < E[i].B[j].size()*(index+1))
					{
					        G[i].B[j][k] = largeScale*randFloat() + largeScale*randFloat()*I;

						G[i].B[j][E[i].B[j].size()*index] = 1.0-largeScale*(randFloat()-2.0)/4.0 + largeScale*(randFloat()-2.0)/4.0*I;
					}
					else
					{
					        G[i].B[j][k] = smallScale*randFloat() + smallScale*randFloat()*I;
					}

					//B[i].B[j][k].real = .40*randFloat();
					//B[i].B[j][k].imag = .40*randFloat();
				}
				++index;
			}
		}
		F = G;
		for(unsigned int i=0; i<F.size(); ++i)
			F[i].orthoganalize();


		dim = saveDim;
		repack();
	}

	// returns q and r s.t. index = div*q + r
	void divAlg(int index, int div, int& q, int& r)
	{
		q = index / div;
		r = index % div;
		if(r < 0)
			r = div + r;
	}
	// finds a and k s.t. index = dim[i]*(a) + k
	// where (a) = (a1,a2,...) = a1+dim[0]*a2 + dim[0]*dim[1]*a3 + ...
	void convertIndexToTuple(int index, std::vector <int>& dimension, std::vector <int>& a, int& k)
	{
		// first k:
		int lambda;
		int D = 1;
		for(unsigned int i=0; i<dimension.size(); ++i)
			D *= dimension[i];

		divAlg(index, D, lambda, k);
		index = lambda;

		D = dimension[0];
		// now a:
		a.resize(dimension.size());
		for(unsigned int i=0; i<dimension.size()-1; ++i)
		{
			int I2 = 0;
			divAlg(index, D, I2, a[i]);
			index = I2;
			D *= dimension[i];
		}
		a[dimension.size()-1] = index;
	}

	// computes (a) = (a0,a1, ...) = a0+dim[0]*a1+dim[0]*dim[1]*a2+...
	int convertTupleToIndex(std::vector <int>& a, std::vector <int>& dimension)
	{
		int D = 1, output = 0;
		for(unsigned int i=0; i<a.size(); ++i)
		{
			output += D*a[i];
			D *= dimension[i];
		}

		return output;
	}

	void clearBasis(std::vector <Complex>& basis)
	{
		for(unsigned int i=0; i<basis.size(); ++i)
		{
		  basis[i] = 0;
		}
	}

	void addToBasis(std::vector <Complex>& output, Complex& scalar, int space, int index, int k)
	{
		std::vector <Complex> temp;
		unsigned int start = k*output.size();
		int j = 0;
		for(unsigned int i=start; i<start+output.size(); ++i)
		{
			//std::cout << space << ", " << index << ", " << i << "\n";
			Complex c = algebra::ComplexHelper::mult(scalar, F[space].B[index][i]);
			output[j] += c;

			++j;
		}
	}

	// given an orthognormal basis F, will "pack" it into a set of elements |e_j>_i
	// it is assumed that all E systems are sequential
	void repack()
	{
		//orig = B;
		std::vector <int> a, dim2,dim3;
		int k, bStart=-1, index;
		//dim2.push_back(2);
		dim2.push_back(dim[0]);
		dim3.push_back(2);
		//dim3.push_back(4);
		for(unsigned int i=0; i<E.size(); ++i)
		{
			//dim2.push_back(dim[i+1]);
			//if(Bsmall[i].B.empty()) continue;
			//if(bStart == -1) bStart = i;
			a.push_back(0);
			//dim.push_back((int)Bsmall[i].B.size());

			for(unsigned int j=0; j<E[i].B.size(); ++j)
			{
				// pack |e_j>_i:
				convertIndexToTuple(j, dim2, a, k);
				int J = a[a.size()-1];
				//convertIndexToTuple(j, dim2, a, k);
				clearBasis(E[i].B[j]);
				if(i == 0)
				{
					Complex one = 1;

					addToBasis(E[i].B[j], one, i, J, k);
				}
				else
				{
					for(int l=0; l < (int)E[i-1].B[J].size(); ++l)
					{
						a[a.size()-1] = l;
						index = convertTupleToIndex(a, dim3);
						//std::cout << J << " " << l << "\n";
						E[i-1].B[J][l];

						addToBasis(E[i].B[j], E[i-1].B[J][l], i, index, k);
					}
				}
			}
			dim2.push_back(dim[i+1]);
			dim3.push_back(4);
			//dim2.push_back((int)Bsmall[i].B.size());
		}
	}

	void crossOver(CS& other, CS& child)
	{
		child.G = G;
		child.F = F;
		child.E = E;

		child.dim = dim;
		child.prodDim = prodDim;
		for(unsigned int i=0; i<G.size(); ++i)
		{
			for(unsigned int j=0; j<G[i].B.size(); ++j)
			{
				int bound = rand()%G[i].B[j].size();
				double p = randFloat();
				p = p + 2;
				p = p / 4.0;
				for(unsigned int k=0; k<G[i].B[j].size(); ++k)
				{
					//if(k < orig[i].B[j].size()/2)
					if(k < bound)
						child.G[i].B[j][k] = G[i].B[j][k];
					else
						child.G[i].B[j][k] = other.G[i].B[j][k];

					p = randFloat();
					p = p + 2;
					p = p / 4.0;
					child.G[i].B[j][k] = p*G[i].B[j][k] + (1-p)*other.G[i].B[j][k];

				}
			}
			child.F[i] = child.G[i];
			child.F[i].orthoganalize();
		}
		child.repack();
	}
	void mutate()
	{
		double stepsize = randFloat()/1.0;//0.1;
		const int rate = 25;
		for(unsigned int i=0; i<G.size(); ++i)
		{
			for(unsigned int j=0; j<G[i].B.size(); ++j)
			{
				for(unsigned int k=0; k<G[i].B[j].size(); ++k)
				{
					stepsize = randFloat()/10.0;
					if(rand()%100 < rate)
					{
						if(rand()%100 < 50)
							G[i].B[j][k] += stepsize;
						else
							G[i].B[j][k] -= stepsize;
					//}
					//if(rand()%100 < rate)
					//{
						stepsize = randFloat()/10.0;
						if(rand()%100 < 50)
							G[i].B[j][k] += stepsize*I;
						else
							G[i].B[j][k] -= stepsize*I;
					}
				}
			}
			F[i] = G[i];
			F[i].orthoganalize();
		}
		repack();
	}
};


class EvolveSystem
{
public:
	void evolve(QKD* protocol, double targetQBER, std::vector <quantum::Basis>& output);
private:
	void fastTrace(algebra::mat& input, algebra::mat& output);

	void rollOver(int& t, int& b, int& a, int& e, int& p);

	double SafeLog(double x)
	{
		if(x <= 0.000000001)
			return 0;
		return log(x)/log(2.0);
	}

	algebra::mat projectAB00, projectAB11, projectAB01, projectAB10;
	algebra::mat BProj0, BProj1, BMeasureOp;
};
}
#endif
